
public class Tree {
	Tree[] children;
	int height;
	float relativePosition, minLeftmostPosition, maxRightmostPosition;
	float[] leftmostPositions, rightmostPositions;
	
	Tree() {
		this.children = new Tree[0];
		height = 1;
		leftmostPositions = new float[1];
		rightmostPositions = new float[1];		
	}
	
	Tree(Tree[] children) {
		this.children = children;
		height = 1;
		for (Tree child : children)
			height = Math.max(height, child.height + 1);
		leftmostPositions = new float[height];
		rightmostPositions = new float[height];
		for (int j = 1; j < height; j++) {
			leftmostPositions[j] = Float.MAX_VALUE;
			rightmostPositions[j] = -Float.MAX_VALUE;
		}
	}
	
	// Mirror symmetry
	Tree mirror() {
		if (height == 1) return new Tree();
		int l = children.length;
		Tree[] mirrorChildren = new Tree[l];
		for (int i = 0; i < l; i++)
			mirrorChildren[i] = children[l-i-1].mirror();
		return new Tree(mirrorChildren);
	}

	// Create a complete regular tree of given degree and height
	static Tree regularTree(int degree, int height) {
		if (height == 1) return new Tree();
		Tree[] children = new Tree[degree];
		for (int i = 0; i < degree; i++)
			children[i] = regularTree(degree, height - 1);
		return new Tree(children);
	}
	
	// Compute all leftmost and rightmost positions, assuming that the relative positions of all children have been chosen
	void updateShadow() {
		for (Tree child : children) {
			for (int j = 0; j < child.height; j++) {
				leftmostPositions[j+1] = Math.min(leftmostPositions[j+1], child.relativePosition + child.leftmostPositions[j]);
				minLeftmostPosition = Math.min(minLeftmostPosition, leftmostPositions[j+1]);
				rightmostPositions[j+1] = Math.max(rightmostPositions[j+1], child.relativePosition + child.rightmostPositions[j]);
				maxRightmostPosition = Math.max(maxRightmostPosition, rightmostPositions[j+1]);
			}
		}
	}
	
	// Compute the maximum value of rightmostPositions[i] - leftmostPositions[j] + 1 
	static float minimalDistance(float[] rightmostPositions, float[] leftmostPositions) {
		float minimalDistance = 1;
		for (int j = 0; j < Math.min(rightmostPositions.length, leftmostPositions.length); j++)
			minimalDistance = Math.max(minimalDistance, rightmostPositions[j] - leftmostPositions[j] + 1);
		return minimalDistance;
	}
	
	// Compute the positions of the children
	void computeChildrenPositions() {
		if (height > 1) {
			float[] leftToRightPositions = new float[height - 1];
			float[] rightToLeftPositions = new float[height - 1];
			leftToRightPositions[0] = -1;
			rightToLeftPositions[0] = 1;
			for (int j = 1; j < height - 1; j++) {
				leftToRightPositions[j] = -Float.MAX_VALUE;
				rightToLeftPositions[j] = Float.MAX_VALUE;
			}
			for (int i = 0; i < children.length; i++) {
				// add next child from left to right
				float leftToRightMinimalDistance = minimalDistance(leftToRightPositions, children[i].leftmostPositions);
				for (int j = 0; j < children[i].height; j++)
					leftToRightPositions[j] = children[i].rightmostPositions[j] + leftToRightMinimalDistance;
				children[i].relativePosition += leftToRightPositions[0] / 2;
				// add next child from right to left
				float rightToLeftMinimalDistance = minimalDistance(children[children.length - i - 1].rightmostPositions, rightToLeftPositions);
				for (int j = 0; j < children[children.length - i - 1].height; j++)
					rightToLeftPositions[j] = children[children.length - i - 1].leftmostPositions[j] - rightToLeftMinimalDistance;
				children[children.length - i - 1].relativePosition += rightToLeftPositions[0] / 2;
			}
		}
	}

	/* Same function separated in two phases
	void computeChildrenPositions() {
		if (height > 1) {
			// add children from left to right
			float[] leftToRightPositions = new float[height - 1];
			for (int j = 1; j < height - 1; j++)
				leftToRightPositions[j] = -Float.MAX_VALUE;
			for (int i = 0; i < children.length; i++) {
				float leftToRightMinimalDistance = minimalDistance(leftToRightPositions, children[i].leftmostPositions);
				for (int j = 0; j < children[i].height; j++)
					leftToRightPositions[j] = children[i].rightmostPositions[j] + leftToRightMinimalDistance;
				children[i].relativePosition = leftToRightPositions[0] / 2;
			}
			// add children from right to left
			float[] rightToLeftPositions = new float[height - 1];
			for (int j = 1; j < height - 1; j++)
				rightToLeftPositions[j] = Float.MAX_VALUE;
			for (int i = children.length - 1; i >= 0; i--) {
				float rightToLeftMinimalDistance = minimalDistance(children[i].rightmostPositions, rightToLeftPositions);
				for (int j = 0; j < children[i].height; j++) {
					rightToLeftPositions[j] = children[i].leftmostPositions[j] - rightToLeftMinimalDistance;
				}
				children[i].relativePosition += rightToLeftPositions[0] / 2;
			}
		}
	}
	*/
	
	/*Functions to pack to the left or to the right
	// Pack the children to the left
	void packChildrenLeft() {
		if (children.length > 0) {
			// add children from left to right
			float[] leftToRightPositions = new float[height - 1];
			leftToRightPositions[0] = -1;
			for (int j = 1; j < height - 1; j++)
				leftToRightPositions[j] = -Float.MAX_VALUE;
			for (int i = 0; i < children.length; i++) {
				float leftToRightMinimalDistance = minimalDistance(leftToRightPositions, children[i].leftmostPositions);
				for (int j = 0; j < children[i].height; j++)
					leftToRightPositions[j] = children[i].rightmostPositions[j] + leftToRightMinimalDistance;
				children[i].relativePosition = leftToRightPositions[0];
			}
			for (Tree child : children)
				child.relativePosition -= leftToRightPositions[0] / 2;
		}
	}

	// Pack the children to the right
	void packChildrenRight() {
		if (children.length > 0) {
			// add children from left to right
			float[] rightToLeftPositions = new float[height - 1];
			rightToLeftPositions[0] = 1;
			for (int j = 1; j < height - 1; j++)
				rightToLeftPositions[j] = Float.MAX_VALUE;
			for (int i = children.length - 1; i >= 0; i--) {
				float rightToLeftMinimalDistance = minimalDistance(children[i].rightmostPositions, rightToLeftPositions);
				for (int j = 0; j < children[i].height; j++)
					rightToLeftPositions[j] = children[i].leftmostPositions[j] - rightToLeftMinimalDistance;
				children[i].relativePosition = rightToLeftPositions[0];
			}
			for (Tree child : children)
				child.relativePosition -= rightToLeftPositions[0] / 2;
		}
	}
	*/
	
	// Compute the planar embedding
	void computeDrawing() {
		for (Tree child : children)
			child.computeDrawing();
		computeChildrenPositions();
		//packChildrenLeft();
		//packChildrenRight();
		updateShadow();
	}
	
	public static void main(String[] args) {
		Tree bt = regularTree(2, 4);
		bt.computeDrawing();
		new TreeViewer(bt);

		Tree tt = regularTree(3, 3);
		tt.computeDrawing();
		new TreeViewer(tt);

		Tree t = new Tree(new Tree[] {new Tree (), new Tree (), new Tree(new Tree[] {new Tree(new Tree[] {new Tree(), new Tree(), new Tree(), new Tree()}), new Tree(), new Tree()}), new Tree(new Tree[] {new Tree(new Tree[] {new Tree(), new Tree(), new Tree(), new Tree()})})});
		t.computeDrawing();
		new TreeViewer(t);
		
		Tree mt = t.mirror();
		mt.computeDrawing();
		new TreeViewer(mt);
	}
}
